use std::ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Neg, Sub, SubAssign};

use crate::{Float, Numeric, Primitive};

/// Struct representing a complex number.
///
/// # Example
///
/// ```rust
/// # use lineal::{Complex, cplx, Float, Numeric};
/// let c = Complex { r: 9.0, i: 0.0 };
/// assert_eq!(format!("{:?}", c), "Complex { r: 9.0, i: 0.0 }");
///
/// let three = c.sqrt().r;
/// assert_eq!(three, 3.0);
///
/// let i: Complex<f32> = cplx!(-1.0).sqrt();
/// assert_eq!(i, Complex::i());
///
/// let a = Complex { r: 2.0, i: 0.5 };
/// let a_squared = a * a;
/// assert_eq!(a_squared, Complex { r: 3.75, i: 2.0 });
/// assert_eq!(a_squared.sqrt(), a);
///
/// let x = (16.0).sqrt();
/// let y = cplx!(16.0).sqrt().r;
/// assert_eq!(x, y);
/// ```
#[derive(Debug, Copy, Clone)]
pub struct Complex<T: Float + Numeric + Primitive> {
    /// The real component of the complex number.
    pub r: T,
    /// The imaginary component of the complex number.
    pub i: T,
}

/// A macro for easier creation of `Complex` numbers.
///
/// # Example
///
/// ```rust
/// # use lineal::{Complex, cplx};
/// // just a real number
/// let a = cplx!(25.0);
/// // first real, second imaginary
/// let b = cplx!(1.0, -1.0);
/// // with named arguments
/// let c = cplx!(r = -1.0,);
/// let d = cplx!(i = 1.0,);
/// let e = cplx!(r = 1.1, i = 0.0,);
/// // even reverse order
/// let f = cplx!(i = 42.0, r = 23.0,);
/// ```
#[macro_export]
macro_rules! cplx {
    ($real:expr, $imag:expr) => {
        {
            Complex { r: $real, i: $imag }
        }
    };
    ($real:expr) => {
        {
            Complex { r: $real, i: 0.0 }
        }
    };
    ($($field:ident = $value:expr),* $(,)?) => {
        {
            Complex {
                $( $field: $value, )*
                ..Complex::default()
            }
        }
    }
}

impl<T: Float + Numeric + Primitive> Default for Complex<T> {
    fn default() -> Self {
        Complex::whole(0)
    }
}

impl<T: Float + Numeric + Primitive> Complex<T> {
    /// Create a complex number with its imaginary component being zero.
    pub fn real(value: T) -> Self {
        Self {
            r: value,
            i: T::whole(0),
        }
    }

    /// Create a complex number with its real component being zero.
    pub fn imag(value: T) -> Self {
        Self {
            r: T::whole(0),
            i: value,
        }
    }

    /// The imaginary unit.
    pub fn i() -> Self {
        Self::imag(T::whole(1))
    }
}

impl<T: Float + Numeric + Primitive> Add for Complex<T> {
    type Output = Self;

    fn add(self, rhs: Self) -> Self::Output {
        Self {
            r: self.r + rhs.r,
            i: self.i + rhs.i,
        }
    }
}

impl<T: Float + Numeric + Primitive> AddAssign for Complex<T> {
    fn add_assign(&mut self, rhs: Self) {
        self.r += rhs.r;
        self.i += rhs.i;
    }
}

impl<T: Float + Numeric + Primitive> Sub for Complex<T> {
    type Output = Self;

    fn sub(self, rhs: Self) -> Self::Output {
        Self {
            r: self.r - rhs.r,
            i: self.i - rhs.i,
        }
    }
}

impl<T: Float + Numeric + Primitive> SubAssign for Complex<T> {
    fn sub_assign(&mut self, rhs: Self) {
        self.r -= rhs.r;
        self.i -= rhs.i;
    }
}

impl<T: Float + Numeric + Primitive> Mul for Complex<T> {
    type Output = Self;

    fn mul(self, rhs: Self) -> Self::Output {
        Self {
            r: self.r * rhs.r - self.i * rhs.i,
            i: self.r * rhs.i + self.i * rhs.r,
        }
    }
}

impl<T: Float + Numeric + Primitive> MulAssign for Complex<T> {
    fn mul_assign(&mut self, rhs: Self) {
        *self = *self * rhs;
    }
}

impl<T: Float + Numeric + Primitive> Div for Complex<T> {
    type Output = Self;

    fn div(self, rhs: Self) -> Self::Output {
        let divisor = rhs.r * rhs.r + rhs.i * rhs.i;
        let mut result = self * Self { r: rhs.r, i: -rhs.i };
        result.r /= divisor;
        result.i /= divisor;
        return result;
    }
}

impl<T: Float + Numeric + Primitive> DivAssign for Complex<T> {
    fn div_assign(&mut self, rhs: Self) {
        *self = *self / rhs;
    }
}

impl<T: Float + Numeric + Primitive> Neg for Complex<T> {
    type Output = Self;

    fn neg(self) -> Self::Output {
        Self {
            r: -self.r,
            i: -self.i,
        }
    }
}

impl<T: Float + Numeric + Primitive> PartialEq for Complex<T> {
    fn eq(&self, other: &Self) -> bool {
        self.r == other.r && self.i == other.i
    }

    fn ne(&self, other: &Self) -> bool {
        self.r != other.r || self.i != other.i
    }
}

impl<T: Float + Primitive + Numeric> Float for Complex<T> {
    fn sqrt(self) -> Self {
        let zero = T::whole(0);
        let two = T::whole(2);
        let abs_z = self.abs().r;

        let base_re = if self.i == zero { self.r } else { (abs_z + self.r) / two };
        let re = if base_re < zero { zero } else { base_re.sqrt() };

        let im = if self.i == zero { zero } else {
            (self.i / self.i.abs()) * ((abs_z - self.r) / two).sqrt()
        } + if base_re < zero { (-base_re).sqrt() } else { zero };

        Self {
            r: re,
            i: im,
        }
    }

    fn sin(self) -> Self {
        Self {
            r: self.r.sin() * self.i.cosh(),
            i: self.r.cos() * self.i.sinh(),
        }
    }

    fn cos(self) -> Self {
        Self {
            r: self.r.cos() * self.i.cosh(),
            i: self.r.sin() * self.i.sinh(),
        }
    }

    fn sinh(self) -> Self {
        Self {
            r: self.r.sinh() * self.i.cos(),
            i: self.r.cosh() * self.i.sin(),
        }
    }

    fn cosh(self) -> Self {
        Self {
            r: self.r.cosh() * self.i.cos(),
            i: self.r.sinh() * self.i.sin(),
        }
    }

    fn value(f: f32) -> Self {
        Self {
            r: T::value(f),
            i: T::whole(0),
        }
    }

    #[cfg(feature = "angular")]
    fn pi() -> Self {
        Self {
            r: T::pi(),
            i: T::whole(0),
        }
    }
}

impl<T: Float + Numeric + Primitive> Numeric for Complex<T> {
    fn whole(value: u8) -> Self {
        Self { r: T::whole(value), i: T::whole(0) }
    }

    fn abs(self) -> Self {
        Self {
            r: (self.r * self.r + self.i * self.i).sqrt(),
            i: T::whole(0),
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::{Complex, Float};

    #[test]
    fn macro_creation() {
        let a = cplx!(-1.0);
        let b = cplx!(r = -1.0,);
        assert_eq!(a, b);
        let c = cplx!(i = 1.0,);
        assert_eq!(a.sqrt(), c);
        let d = cplx!(r = -1.0, i = 0.0,);
        assert_eq!(c * c, d);
        let e = cplx!(-1.0, 0.0);
        assert_eq!(a, e);
    }

    #[test]
    fn division() {
        let mut a: Complex<f32> = Complex { r: 3.0, i: 2.0 };
        let b: Complex<f32> = Complex { r: 4.0, i: -5.0 };
        let expected: Complex<f32> = Complex { r: 2.0 / 41.0, i: 23.0 / 41.0 };
        a /= b;
        assert_eq!(format!("{:?}", a), format!("{:?}", expected));
    }

    #[test]
    fn square_root_with_negatives() {
        let a = Complex { r: -1.0, i: -1.0 };
        let root: Complex<f64> = a.sqrt();
        let expected = Complex { r: 0.45508986, i: -1.09868411 };
        assert!((root.r - expected.r).abs() < 0.00001);
        assert!((root.i - expected.i).abs() < 0.00001);
    }
}
